/*
 *  Copyright 2016 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <common.h>
#include <bootm.h>
#include <command.h>
#include <linux/sizes.h>
#include <fs.h>
#include <malloc.h>
#include <u-boot/crc.h>
#include <boot_time.h>
#include <part.h>
#include <log.h>
#include <asm/cache.h>
#include <asm/unaligned.h>
#include <image.h>

#include "ssboot.h"

#define FS_READ_SIZE SZ_16M

/*
 * Reserved for all functions of u-boot except SBI temporary buffer.
 * Such as filesystems，drivers，and other commands.
 */
#define MALLOC_RESERVED_SIZE	SZ_4M

/* Extra malloc space will be needed when loading an LZ4-compressed
 * snapshot as the complete fill will be loaded at once into the memory.
 */
#if (CONFIG_SYS_MALLOC_LEN < FS_READ_SIZE + MALLOC_RESERVED_SIZE)
#error "ssboot: try increasing CONFIG_SYS_MALLOC_LEN value"
#endif


#ifdef CONFIG_CMD_SSBOOT_TIME
#define FLOAT(size, n) (((((size) >> (n)) & 0x3FF) * 1000) / 1024)

static void show_speed(unsigned long size, unsigned long time)
{
	unsigned long speed;
	unsigned long ms;

	if (size >> 30)
		printf("Size : %lu.%03luGB\n", size >> 30, FLOAT(size, 20));
	else if (size >> 20)
		printf("Size : %lu.%03luMB\n", size >> 20, FLOAT(size, 10));
	else if (size >> 10)
		printf("Size : %lu.%03luKB\n", size >> 10, FLOAT(size, 0));
	else
		printf("Size : %luB\n", size);
	ms = time / (CONFIG_SYS_HZ / 1000);
	if (ms > 1000)
		printf("Time : %lu.%03lus\n", ms / 1000, ms % 1000);
	else
		printf("Time : %lums\n", ms);

	speed = (size / ms) * 1000;
	if (speed >> 30)
		printf("Speed: %lu.%03luGB/s\n", speed >> 30, FLOAT(speed, 20));
	else if (speed >> 20)
		printf("Speed: %lu.%03luMB/s\n", speed >> 20, FLOAT(speed, 10));
	else if (speed >> 10)
		printf("Speed: %lu.%03luKB/s\n", speed >> 10, FLOAT(speed, 0));
	else
		printf("Speed: %luB/s\n", speed);
}
#endif

static loff_t exec_fs_size(struct ss_ctx *ctx)
{
	loff_t filesize;

	if (fs_set_blk_dev(ctx->interface, ctx->devpart, ctx->fs_type)) {
		printf("Error: setting partition %s on interface %s failed\n",
				ctx->devpart, ctx->interface);
		return -1;
	}

	if (fs_size(ctx->filename, &filesize) < 0) {
		printf("Error: cannot get file size\n");
		return -1;
	}

	return filesize;
}

static loff_t get_elf_size(struct ss_ctx *ctx, Elf_Phdr phdr)
{
	loff_t partition_size, filesize;
	struct blk_desc *fs_dev_desc;
	struct disk_partition fs_partition;

	if (blk_get_device_part_str(ctx->interface, ctx->devpart, &fs_dev_desc,
		&fs_partition, 1) < 0)
		return -1;

	partition_size = fs_partition.size * fs_partition.blksz;

	filesize = exec_fs_size(ctx);
	if (filesize <= 0)
		return filesize;

	if (filesize < phdr.p_filesz + phdr.p_offset) {
		printf("Error: ssboot image has been truncated.\n");
		return -1;
	}

	/* is rawfs */
	if (partition_size == filesize) {
		filesize = phdr.p_filesz + phdr.p_offset;

		if (filesize <= 0 || filesize > partition_size) {
			filesize = partition_size;
			printf("Warning: ELF size is invalid,"
					"try using the partition size\n");
		}
	}

	return filesize;
}

/*
 * Decompress 'n' lz4 blocks from comp. buffer and put the result in elf buffer.
 */
static int decompress_lz4(int n, struct ss_ctx *ctx)
{
	int ret = 0;
	void *in   = buf_cur(&ctx->cbuf);
	void *iend = ctx->cbuf.data + ctx->cbuf.size;
	void *out  = buf_cur(&ctx->ebuf);
	void *oend = ctx->ebuf.data + ctx->ebuf.size;

	while (n--) {
		u32 block_header, block_size;

		block_header = get_unaligned_le32(in);
		in += sizeof(u32);
		block_size = block_header & ~LZ4F_BLOCKUNCOMPRESSED_FLAG;

		if (in + block_size > iend) {
			ret = -EINVAL;
			break;
		}
		if (!block_size) {
			ret = 0;
			break;
		}
		if (block_header & LZ4F_BLOCKUNCOMPRESSED_FLAG) {
			size_t size = min((ptrdiff_t)block_size, oend - out);

			memcpy(out, in, size);
			out += size;
			if (size < block_size) {
				ret = -ENOBUFS;
				break;
			}
		} else {
			/* constant folding essential, do not touch params! */
			ret = LZ4_decompress_generic(in, out, block_size,
					oend - out, endOnInputSize, full, 0,
					noDict, out, NULL, 0);
			if (ret < 0) {
				ret = -EPROTO;
				break;
			}
			out += ret;
			debug("LZ4 block decoded: size %d B -> %d B\n",
					block_size, ret);
		}
		in += block_size;
		if (ctx->has_block_checksum)
			in += sizeof(u32);
	}
	ctx->cbuf.start += in - buf_cur(&ctx->cbuf);
	ctx->ebuf.size  = out - buf_cur(&ctx->ebuf);

	return ret;
}

/*
 * Fill the ELF buffer with data. Called only when the buffer is empty.
 */
static loff_t fill_elf_buffer(struct ss_ctx *ctx)
{
	int res;

	if (ctx->mode == SSBOOT_ELF_FILE) {
		loff_t len, act_read_len;
		struct ss_buf *buf;

		if (fs_set_blk_dev(ctx->interface, ctx->devpart, ctx->fs_type)) {
			printf("Error: setting partition %s on interface %s failed\n",
					ctx->devpart, ctx->interface);
			return -1;
		}

		buf = &ctx->ebuf;
		if (ctx->file_total_size < ctx->total_read_len + buf->size) {
			len = ctx->file_total_size - ctx->total_read_len;
			buf->size = len;
		} else {
			len = buf->size;
		}
		/* The entire ssboot file has been loaded, but more data is needed */
		if (len == 0)
			return -1;

		res = fs_read(ctx->filename, (ulong)buf->data, ctx->total_read_len,
				len, &act_read_len);
		if (res < 0 || act_read_len != len) {
			printf("Error: read of %lld bytes from %s:%lld failed\n",
					len, ctx->filename, ctx->total_read_len);
			return -1;
		}
		ctx->total_read_len += act_read_len;
		buf->start = 0;
		buf->size  = act_read_len;
		res = act_read_len;

	} else if (ctx->mode == SSBOOT_LZ4_MEM) {
		ctx->ebuf.start = 0;
		res = decompress_lz4(4, ctx);
		if (res < 0)
			printf("Error: LZ4 decompression failed!\n");
	}

	return res;
}

/*
 * Return
 * 0 : all data in buf is relocated
 * >0: some bytes data is not relocated
 * -1: error
 */

#ifdef CONFIG_CMD_SSBOOT_CKSUM
static unsigned int inflight_data_cksum;
#endif

static loff_t elf_load_sect(const Elf_Phdr *phdr, struct ss_ctx *ctx)
{
	loff_t ret;
	void *dst;
	loff_t filesz, blen;
	struct ss_buf *buf = &ctx->ebuf;

	filesz = phdr->p_filesz;
	dst = (void *)phdr->p_paddr;

	while (filesz > 0) {
		if (buf_len(buf) == 0) {
			/* load new data */
			ret = fill_elf_buffer(ctx);
			if (ret < 0)
				return -1;
		}
		blen = buf_len(buf);
		if (blen >= filesz) {
			debug("dst=%p src=%p, filesz=0x%llx\n", dst, buf_cur(buf), filesz);
#ifdef CONFIG_CMD_SSBOOT_CKSUM
			inflight_data_cksum = crc32(inflight_data_cksum, buf_cur(buf), filesz);
#ifdef CONFIG_CMD_SSBOOT_DEBUG_CORRUPTION
			printf("load %p -> %p: %08x\n", (void *) phdr->p_paddr,
					(void *) (phdr->p_paddr + phdr->p_filesz),
					inflight_data_cksum);
#endif
#endif /* CONFIG_CMD_SSBOOT_CKSUM */
			memcpy(dst, buf_cur(buf), filesz);
			buf->start += filesz;
			/* completed, turn to next */
			break;
		}
		/* not completed, copy all */
#ifdef CONFIG_CMD_SSBOOT_CKSUM
		inflight_data_cksum = crc32(inflight_data_cksum, buf_cur(buf), blen);
#endif
		memcpy(dst, buf_cur(buf), blen);
		debug("dst=%p src=%p, rest_size=0x%llx\n", dst, buf_cur(buf), blen);
		filesz -= blen;
		dst += blen;
		buf->start = buf->size; // buf_len = 0
	}

	return 0;
}

#ifdef CONFIG_CMD_SSBOOT_CKSUM
static unsigned int ssboot_calc_data_checksum(Elf_Ehdr *ehdr, Elf_Phdr *phdr)
{
	int i;
	unsigned int data_cksum = 0;

	for (i = 0; i < ehdr->e_phnum; i++, phdr++) {
		if (phdr->p_type != PT_LOAD)
			continue;

		if (phdr->p_filesz == 0)
			continue;

		data_cksum = crc32(data_cksum, (void *)phdr->p_paddr,
				   phdr->p_filesz);
#ifdef CONFIG_CMD_SSBOOT_DEBUG_CORRUPTION
		printf("check %p -> %p: %08x\n", (void *) phdr->p_paddr,
				(void *) (phdr->p_paddr + phdr->p_filesz),
				data_cksum);
#endif
	}

	return data_cksum;
}

static int ssboot_verify_checksum(Elf_Ehdr *ehdr, Elf_Phdr *phdr)
{
	int i;
	Elf_Phdr *cksum_phdr = NULL;
	unsigned int ehdr_cksum, phdr_cksum, data_cksum;

	for (i = 0; i < ehdr->e_phnum; i++) {
		if (phdr[i].p_type != PT_LOAD) {
			if (phdr[i].p_offset == CKSUM_TYPE_CRC32)
				cksum_phdr = &phdr[i];
			continue;
		}
		break;
	}

	if (NULL == cksum_phdr) {
		printf("Warning: checksum not found in image, skipping verification\n");
		return 0;
	}

	ehdr_cksum = crc32(0, (void *)ehdr, sizeof(*ehdr));
	phdr_cksum = crc32(0, (void *)&cksum_phdr[1],
			ehdr->e_phnum * ehdr->e_phentsize
			- ((size_t)&cksum_phdr[1] - (size_t)phdr));
	data_cksum = ssboot_calc_data_checksum(ehdr, phdr);

	if (elf_ehdr_cksum(cksum_phdr) != ehdr_cksum ||
		elf_phdr_cksum(cksum_phdr) != phdr_cksum ||
		elf_data_cksum(cksum_phdr) != data_cksum) {
		printf("ehdr_cksum: %08x %08x\n",
			(unsigned int) elf_ehdr_cksum(cksum_phdr), ehdr_cksum);
		printf("phdr_cksum: %08x %08x\n",
			(unsigned int) elf_phdr_cksum(cksum_phdr), phdr_cksum);
		printf("data_cksum: %08x %08x (during transfer %08x)\n",
			(unsigned int) elf_data_cksum(cksum_phdr), data_cksum, inflight_data_cksum);

		if(elf_data_cksum(cksum_phdr) != inflight_data_cksum)
			printf("snapshot image is probably corrupt\n");
		if(elf_data_cksum(cksum_phdr) != data_cksum)
			printf("snapshot data was corrupted during load\n");

		return -1;
	}

	return 0;
}
#endif

static int get_elf_headers(struct ss_elf *elf, void *addr)
{
	loff_t phdrs_size;

	memcpy(&elf->ehdr, addr, sizeof(Elf_Ehdr));
	phdrs_size = elf->ehdr.e_phnum * elf->ehdr.e_phentsize;

	elf->phdrs = malloc(phdrs_size);
	if (!elf->phdrs) {
		printf("Error: cannot allocate %llu B memory for program headers!\n",
				phdrs_size);
		return -1;
	}

	memcpy(elf->phdrs, addr + elf->ehdr.e_phoff, phdrs_size);

	return 0;
}

/*
 * Read and load ELF data into memory from a buffer.
 * This buffer is backed by an ELF file on the disk.
 */
static int read_elf_from_buffer(struct ss_ctx *ctx)
{
	int i, ret;
	Elf_Ehdr *ehdr;
	Elf_Phdr *phdrs;
	struct ss_buf *buf = &ctx->ebuf;

	ret = fill_elf_buffer(ctx);
	if (ret < 0)
		return -1;

	ret = get_elf_headers(&ctx->elf, buf->data);
	if (ret < 0)
		return -1;

	ehdr = &ctx->elf.ehdr;
	phdrs = ctx->elf.phdrs;

	if (buf->size < ehdr->e_phoff + ehdr->e_phnum * ehdr->e_phentsize) {
		printf("Error: temp buffer is too small for ELF header info\n");
		return -1;
	}

	/* the ELF file size can be calculated only after the program headers are read */
	if (ctx->mode == SSBOOT_ELF_FILE) {
		ctx->file_total_size = get_elf_size(ctx, phdrs[ehdr->e_phnum - 1]);
		if (ctx->file_total_size < 0) {
			free(ctx->elf.phdrs);
			return -1;
		}
	}

	for (i = 0; i < ehdr->e_phnum; i++) {
		if (phdrs[i].p_type != PT_LOAD)
			continue;
		break;
	}
	buf->start = phdrs[i].p_offset;
	for (; i < ehdr->e_phnum; i++) {
		if (phdrs[i].p_type != PT_LOAD)
			continue;
		if (elf_load_sect(phdrs + i, ctx)) {
			printf("Error: load of section %d failed\n", i);
			free(ctx->elf.phdrs);
			return -1;
		}
		if (phdrs[i].p_filesz != phdrs[i].p_memsz)
			memset((void *)phdrs[i].p_paddr + phdrs[i].p_filesz, 0x00,
					phdrs[i].p_memsz - phdrs[i].p_filesz);
	}

	return 0;
}

/*
 * LZ4 format does not store the total compressed data size.
 * Calculate by summing up individual blocks' size.
 */
static loff_t get_lz4_file_size(void *in, int has_block_cksum, int max_block_size)
{
	u32 block_header, block_size;
	void *cur = in;

	while (1) {
		block_header = get_unaligned_le32(cur);
		block_size = block_header & ~LZ4F_BLOCKUNCOMPRESSED_FLAG;
		if (block_size > max_block_size)
			return 0;
		if (block_size == 0) {
			cur += sizeof(u32);
			break;
		}
		cur += sizeof(u32) + block_size + (has_block_cksum? sizeof(u32): 0);
	}
	return cur - in;
}

/* Load ELF data from an lz4-compressed snapshot file loaded in memory */
static int load_lz4_from_mem(struct ss_ctx *ctx, void *addr)
{
	int ret;
	loff_t lz4_file_size;
	u32 magic;
	u8 flags, block_desc;
	u8 version, independent_blocks;
	u8 has_content_size, has_block_checksum;
	u8 max_block_size;
	const int lz4_block_sizes[] = {SZ_64K, SZ_256K, SZ_1M, SZ_4M};
	void *in = addr;

	printf("## Loading LZ4-compressed snapshot from address 0x%08lx\n",
			(ulong)addr);

	magic		= get_unaligned_le32(addr);
	addr		+= sizeof(u32);
	flags		= *(u8 *)addr;
	addr		+= sizeof(u8);
	block_desc	= *(u8 *)addr;
	addr		+= sizeof(u8);

	version			= LZ4F_VERSION(flags);
	independent_blocks	= LZ4F_BLOCK_INDEP(flags);
	has_block_checksum	= LZ4F_HAS_BLOCK_CKSUM(flags);
	has_content_size	= LZ4F_HAS_CONTENT_SIZE(flags);
	max_block_size		= LZ4F_MAX_BLOCK_SIZE(block_desc);

	if (magic != LZ4F_MAGIC || version != 1 || independent_blocks != 1)
		return -EPROTONOSUPPORT;
	if ((flags & LZ4F_FLAGS_RESERVED_MASK) ||
			(block_desc & LZ4F_BLOCK_DESC_RESERVED_MASK))
		return -EINVAL;
	ctx->has_block_checksum = has_block_checksum;

	if (has_content_size)
		addr += sizeof(u64);
	addr += sizeof(u8);

	if (max_block_size < 4 || max_block_size > 7)
		return -EINVAL;
	ctx->max_block_size = lz4_block_sizes[max_block_size - 4];

	lz4_file_size = addr - in;
	ret = get_lz4_file_size(addr, ctx->has_block_checksum,
			ctx->max_block_size);
	if (ret == 0) {
		printf("Error: cannot determine lz4 file size!\n");
		return -1;
	}
	lz4_file_size += ret;
	ctx->file_total_size = lz4_file_size;

	ctx->cbuf.data = malloc(lz4_file_size);
	if (!ctx->cbuf.data) {
		printf("Error: cannot allocate %llu B memory!\n", lz4_file_size);
		return -ENOMEM;
	}
	ctx->cbuf.size  = lz4_file_size;
	ctx->cbuf.start = 0;
	memcpy(ctx->cbuf.data, addr, lz4_file_size);

	ctx->ebuf.data = malloc(FS_READ_SIZE);
	if (!ctx->ebuf.data) {
		printf("Error: cannot allocate %d B memory!\n", FS_READ_SIZE);
		free(ctx->cbuf.data);
		return -ENOMEM;
	}
	ctx->ebuf.size  = FS_READ_SIZE;
	ctx->ebuf.start = 0;

	ret = read_elf_from_buffer(ctx);
	if (ret < 0)
		printf("Error: loading ELF data from buffer failed!\n");

	free(ctx->cbuf.data);
	free(ctx->ebuf.data);

	return ret;
}

/* Load ELF data from a snapshot file loaded in memory */
static int load_elf_from_mem(struct ss_ctx *ctx, void *addr)
{
	int i, ret;
	loff_t elf_file_size;
	Elf_Ehdr *ehdr;
	Elf_Phdr *phdrs;
	void *buf;

	printf("## Loading ELF snapshot from address 0x%08lx\n", (ulong)addr);

	ret = get_elf_headers(&ctx->elf, addr);
	if (ret < 0)
		return -1;

	ehdr = &ctx->elf.ehdr;
	phdrs = ctx->elf.phdrs;

	elf_file_size = phdrs[ehdr->e_phnum - 1].p_filesz
			+ phdrs[ehdr->e_phnum - 1].p_offset;
	ctx->file_total_size = elf_file_size;

	buf = malloc(elf_file_size);
	if (!buf) {
		printf("Error: cannot allocate %llu B memory!\n", elf_file_size);
		return -1;
	}

	memcpy(buf, addr, elf_file_size);
	ehdr = (Elf_Ehdr *)buf;
	phdrs = (Elf_Phdr *)(buf + ehdr->e_phoff);
	for (i = 0; i < ehdr->e_phnum; i++) {
		if (phdrs[i].p_type != PT_LOAD)
			continue;
		memcpy((void *)phdrs[i].p_paddr, (void *)(buf + phdrs[i].p_offset),
				phdrs[i].p_filesz);
		if (phdrs[i].p_filesz != phdrs[i].p_memsz)
			memset((void *)phdrs[i].p_paddr + phdrs[i].p_filesz, 0x00,
					phdrs[i].p_memsz - phdrs[i].p_filesz);
	}
	free(buf);

	return 0;
}

/* Load ELF data from a snapshot file on the disk */
static int load_elf_from_file(struct ss_ctx *ctx)
{
	int ret;

	printf("## Loading ELF snapshot from disk\n");

	ctx->ebuf.data = memalign(ARCH_DMA_MINALIGN, FS_READ_SIZE);
	if (!ctx->ebuf.data) {
		printf("Error: cannot allocate %d B memory!\n", FS_READ_SIZE);
		return -1;
	}
	ctx->ebuf.size = FS_READ_SIZE;
	ctx->ebuf.start = 0;
	ctx->file_total_size = FS_READ_SIZE;

	ret = read_elf_from_buffer(ctx);
	if (ret < 0)
		printf("Error: loading ELF data from buffer failed!\n");
	free(ctx->ebuf.data);

	return ret;
}

static ulong load_from_fit_image(void *addr, const char *uname, ulong *data, ulong *len)
{
	int		verify;
	const void	*fit_hdr;
	int		noffset;
	const void	*fit_data;
	size_t		fit_len;
	uint8_t		comp;

	verify = env_get_yesno("verify");

	printf("## Copying '%s' subimage from FIT image "
			"at 0x%08lx ...\n", uname, (ulong)addr);

	fit_hdr = (const void *)addr;
	if (fit_check_format(fit_hdr, IMAGE_SIZE_INVAL)) {
		printf("Error: bad FIT image format\n");
		return -1;
	}

	/* get subimage node offset */
	noffset = fit_image_get_node(fit_hdr, uname);
	if (noffset < 0) {
		printf("Error: can't find '%s' FIT subimage\n", uname);
		return -1;
	}

	/* verify integrity */
	if (verify) {
		if (!fit_image_verify(fit_hdr, noffset)) {
			printf("Error: bad data hash\n");
			return -1;
		}
	}

	/* get subimage/external data address and length */
	if (fit_image_get_data_and_size(fit_hdr, noffset,
				       &fit_data, &fit_len)) {
		puts("Error: could not find script subimage data\n");
		return -1;
	}

	if (fit_image_get_comp(fit_hdr, noffset, &comp)) {
		puts("Error: could not find script subimage compression type\n");
		return -1;
	}

	*data = (ulong)fit_data;
	*len = (ulong)fit_len;

	return 0;
}

static int
do_ssbootelf(struct cmd_tbl *cmdtp, int flag, int argc, char * const argv[])
{
	int ret;
#ifdef CONFIG_ELF64
	Elf64_Addr entry;
#else
	Elf32_Addr entry;
#endif
	void *addr;
	char *ep;
#ifdef CONFIG_CMD_SSBOOT_TIME
	unsigned long time;
#endif
	struct ss_ctx ctx;

	if (argc < 2 || argc > 4)
		return CMD_RET_USAGE;

#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(0);
#endif

	BOOT_TIME_ADD("U-Boot: ssbootelf start");

	addr = (void *)(simple_strtoul(argv[1], &ep, 16));
	if (ep == argv[1] || *ep != '\0') {
		/* first argument is not an address, consider it as a disk interface */
		if (argc < 3 || argc > 4)
			return CMD_RET_USAGE;

		ctx.mode = SSBOOT_ELF_FILE;
		ctx.interface = argv[1];
		ctx.devpart = argv[2];
		if (argc == 4) {
			ctx.fs_type = SSBOOT_FS_TYPE;
			ctx.filename = argv[3];
		} else {
			ctx.fs_type = FS_TYPE_RAW;
			ctx.filename = "raw data";
		}
		ctx.total_read_len = 0;
		if (fs_set_blk_dev(ctx.interface, ctx.devpart, ctx.fs_type)) {
			printf("Error: setting partition %s on interface %s failed!\n",
			ctx.devpart, ctx.interface);
			goto err;
		}
	} else {
		/* first argument is an address */
		Elf_Ehdr *ehdr = (Elf_Ehdr *)addr;
		u32 lz4f_magic = get_unaligned_le32(addr);
		ulong data, len = 0;
		const char *image_uname;

		switch (argc) {
		case 3: /* FIT image loaded in memory */
			if (genimg_get_format((void *)addr) == IMAGE_FORMAT_FIT) {
				image_uname = argv[2];
			} else {
				printf("Error: no valid FIT image found at address %08lx!\n",
					(ulong)addr);
				return CMD_RET_FAILURE;
			}
			ret = load_from_fit_image(addr, image_uname, &data, &len);
			if (ret < 0)
				return CMD_RET_FAILURE;
			else
				addr = (void *)data;
		/* -- falling-through -- */
		case 2: /* ELF/LZ4 snapshot loaded in memory */
			ehdr = (Elf_Ehdr *)addr;
			lz4f_magic = get_unaligned_le32(addr);

			if (IS_ELF(*ehdr) && ehdr->e_type == ET_EXEC) {
				ctx.mode = SSBOOT_ELF_MEM;
				ctx.image_size = len;
			} else if (lz4f_magic == LZ4F_MAGIC) {
				ctx.mode = SSBOOT_LZ4_MEM;
				ctx.image_size = len;
			} else {
				printf("Error: no valid ELF or LZ4 file found at address %08lx!\n",
						(ulong)addr);
				return CMD_RET_FAILURE;
			}
			break;
		default:
			break;
		}
	}

	switch (ctx.mode) {
	case SSBOOT_LZ4_MEM:
		ret = load_lz4_from_mem(&ctx, addr);
		break;
	case SSBOOT_ELF_FILE:
		ret = load_elf_from_file(&ctx);
		break;
	case SSBOOT_ELF_MEM:
		ret = load_elf_from_mem(&ctx, addr);
		break;
	default:
		return -1;
	}
	if (ret < 0)
		goto err;


	BOOT_TIME_ADD("U-Boot: ssbootelf end");

#ifdef CONFIG_CMD_SSBOOT_CKSUM
	inflight_data_cksum = 0;

	if (ssboot_verify_checksum(&ctx.elf.ehdr, ctx.elf.phdrs) < 0) {
		printf("Error: checksum verification failed\n");
		goto err;
	}
#endif
	entry = ctx.elf.ehdr.e_entry;
	free(ctx.elf.phdrs);

#ifdef CONFIG_CMD_SSBOOT_TIME
	time = get_timer(time);
	show_speed(ctx.file_total_size, time);
#endif

	BOOT_TIME_ADD("U-Boot: boot_jump_ssboot");
	printf("## Booting snapshot image at 0x%p ...\n", (void *)entry);

	cleanup_before_linux();
	/*
	 * Please don't do anything between cleanup_before_linux() and jump to boot
	 * entry. Because cleanup_before_linux() disabled all of the caches.
	 */
	((void (*)(void))entry)();

	return CMD_RET_SUCCESS;

err:
	free(ctx.elf.phdrs);
	return CMD_RET_FAILURE;
}

U_BOOT_CMD(
	ssbootelf,      4,      0,      do_ssbootelf,
	"Load ssboot ELF image and boot",
	"<interface> <dev:part> [filename]\n"
	"\t- Load ssboot ELF image 'filename' from 'dev:part' on 'interface'"
	"<addr>\n"
	"\t- Load ssboot ELF or LZ4 image from the memory address 'addr'\n"

);
